/**
 * @see displayName
 * @see require
 */
interface FocusTargetOptions {
    /**
     * A human-readable name for debugging focus targets.
     */
    displayName?: string;
    /**
     * Requires the focus target to be active before the coroutine can resume.
     * Instead of the default behaviour, where the coroutine also resumes if no focus target is active.
     */
    require?: boolean;
}

declare const FOCUS_TARGET_OPTIONS_KEY: unique symbol;

/**
 * @see focus
 * @see blur
 */
interface FocusTarget {
    /**
     * The options passed to `createFocusTarget`.
     *
     * @internal
     */
    [FOCUS_TARGET_OPTIONS_KEY]: FocusTargetOptions;
    /**
     * Activates this focus target, allowing its coroutines to run.
     * As only one focus target can be active at a time, this also deactivates every other focus target.
     */
    focus(): void;
    /**
     * Deactivates this focus target, allowing every coroutine to run.
     *
     * If this focus target was created with `require: true`, any coroutines using it will pause.
     */
    blur(): void;
}

interface ControllerManagerInterface$1 {
    focus(target: FocusTarget): void;
    blur(target: FocusTarget): void;
    registerCombinedFocusTarget(target: FocusTarget, sources: readonly FocusTarget[]): void;
}

declare class FocusTargetManager {
    #private;
    constructor(controllerInterface: ControllerManagerInterface$1);
    createFocusTarget(options?: FocusTargetOptions): FocusTarget;
    createCombinedFocusTarget(sources: readonly FocusTarget[]): FocusTarget;
}

/**
 * The value returned from a `yield` statement in a coroutine.
 */
interface CoroutineContext<Context> {
    /**
     * Per-step context from the manager.
     */
    ctx: Context;
    /**
     * Awaiter result data.
     */
    data: unknown;
}

/**
 * Exotic awaiters have special behaviour when `yield`ed.
 *
 * Their fields are internal and subject to change.
 */
interface ExoticCoroutineAwait<Marker extends symbol> {
    marker: Marker;
}

interface CoroutineHandlerOptions<Context> {
    /**
     * A debug identifier for this coroutine.
     */
    identifier: string;
    /**
     * The coroutine function itself.
     */
    coroutine: CoroutineGeneratorFunction<Context>;
}
interface CoroutineHandlerStartInterface {
    /**
     * The parent coroutine's stack trace.
     */
    parentStackTrace: readonly string[];
    /**
     * Stops the coroutine from being updated by removing it from the active list.
     */
    delete(): void;
    /**
     * Checks if the passed focus target is currently focused.
     */
    isFocused(focusTarget: FocusTarget): boolean;
}
declare class Coroutine<Context> {
    #private;
    constructor({ identifier, coroutine }: CoroutineHandlerOptions<Context>);
    start(startInterface: CoroutineHandlerStartInterface): void;
    dispose(): void;
    handle(ctx: Context): void;
    /**
     * Returns the identifier of this coroutine.
     *
     * Useful for displaying debug information.
     */
    getIdentifier(): string;
    /**
     * Returns the number of `shouldContinue` calls that ran on the last `handle` call.
     *
     * Useful for displaying debugging information.
     */
    getLastCheckCount(): number;
}

type CoroutineAwaitResult_Done<Data> = Data extends void | never | undefined ? {
    done: true;
} : {
    done: true;
    data: Data;
};
type CoroutineAwaitResult_Wait = {
    done: false;
    data?: undefined;
};
type CoroutineAwaitResult<Data> = CoroutineAwaitResult_Done<Data> | CoroutineAwaitResult_Wait;

interface CoroutineAwait<Context, Data> {
    /**
     * # Remember to `yield` this call!
     * You may get unexpected bugs if you don't.
     */
    ["Did you forget `yield`?"]?: never;
    /**
     * Identifies the awaiter, used for debugging.
     */
    identifier: string;
    /**
     * When defined, this awaiter is skipped until this promise resolves. Rejection is ignored.
     */
    delay?: Promise<void>;
    /**
     * The focus target to use to control this awaiter.
     *
     * @see FocusTargetManager.createFocusTarget
     */
    focusTarget?: FocusTarget;
    shouldContinue(ctx: Context): CoroutineAwaitResult<Data>;
}

/**
 * Each type that can be automatically casted to an awaiter.
 */
type AwaiterCastable<Context> = CoroutineAwait<Context, unknown> | Coroutine<Context> | CoroutineGeneratorFunction<Context> | undefined;

type CoroutineGeneratorYieldable<Context> = AwaiterCastable<Context> | ExoticCoroutineAwait<symbol>;
type CoroutineGenerator<Context> = Generator<CoroutineGeneratorYieldable<Context>, void, CoroutineContext<Context>>;

type CoroutineGeneratorFunction<Context> = () => CoroutineGenerator<Context>;

interface StartCoroutineResult {
    /**
     * Disposes of and stops this coroutine.
     */
    stop(): void;
}

interface CommonAwaiterOptions {
    /**
     * The focus target to use to control this awaiter.
     * @see FocusTargetManager.createFocusTarget
     */
    focusTarget?: FocusTarget;
}

interface ControllerManagerInterface<Context> {
    /**
     * Registers a coroutine to run each tick.
     */
    registerCoroutine(coroutine: Coroutine<Context>): void;
    /**
     * Stops a coroutine from running each tick.
     *
     * Throws an error if the component isn't registered.
     */
    deregisterCoroutine(coroutine: Coroutine<Context>): void;
    /**
     * Called when the active focus target changes, to store the identifier for debugging.
     */
    handleActiveFocusTargetChanged(identifier: string | null | undefined): void;
}

/**
 * Coroutine controls
 *
 * @param Context The type of the `ctx` property in the coroutine result.
 */
declare class CoroutineManager<Context> {
    #private;
    constructor(controllerInterface: ControllerManagerInterface<Context>);
    startCoroutine(coroutine: Coroutine<Context> | CoroutineGeneratorFunction<Context>): StartCoroutineResult;
    getFocusTargetManager(): FocusTargetManager;
    /**
     * When yielded, registers the callback to be called when this coroutine is about to be disposed.
     * This is always during a `yield` statement (which will never complete).
     */
    hookDispose(callback: () => void): ExoticCoroutineAwait<symbol>;
    /**
     * When yielded, sets the default awaiter options for a coroutine, for any calls after this one.
     * Calling multiple times shallow merges the options.
     */
    hookOptions(options: CommonAwaiterOptions): ExoticCoroutineAwait<symbol>;
    isFocusGlobal(): boolean;
}

declare class CoroutineController<Context> {
    #private;
    getManager(): CoroutineManager<Context>;
    /**
     * Runs all coroutines.
     * @param ctx Context returned as `.ctx` from any `yield`s.
     */
    tick(ctx: Context): void;
    /**
     * Returns the number of coroutines currently registered.
     *
     * Useful for showing debug information.
     */
    getCoroutineCount(): number;
    /**
     * Returns the number of `shouldContinue` calls that ran in the last `tick()` call.
     *
     * Useful for showing debug information.
     */
    getLastCheckCount(): number;
    /**
     * Returns the number of coroutines that were disposed in the last `tick()` call.
     *
     * Useful for showing debug information.
     */
    getDisposedCountThisTick(): number;
    getRegisteredIdentifiers(): string[];
    getActiveFocusTargetIdentifier(): string | null | undefined;
}

interface NestOptions<Context, Result> {
    /**
     * A debug identifier for this nest call.
     *
     * If `null`, no stack trace item is added for the nest call.
     */
    identifier: string | null;
    /**
     * The list of awaiters to nest.
     */
    awaiters: readonly AwaiterCastable<Context>[];
    /**
     * A method that is called once per tick to handle the current results of the awaiters.
     * @param ctx The coroutine context.
     * @param results A list of the results from each awaiter, in the same order as passed.
     */
    handler(ctx: Context, results: CoroutineAwaitResult<unknown>[]): CoroutineAwaitResult<Result>;
}
type NextAwait = ExoticCoroutineAwait<symbol>;
declare function nest<Context, Result>(options: NestOptions<Context, Result>): NextAwait;

/**
 * Waits until one of the awaiters finishes, then returns the index of that awaiter.
 */
declare function one<Context>(awaiters: readonly AwaiterCastable<Context>[]): NextAwait;

/**
 * Waits until every awaiter has finished.
 */
declare function all<Context>(awaiters: readonly AwaiterCastable<Context>[]): NextAwait;

/**
 * An awaiter that waits until the next tick.
 *
 * For example, if the coroutine is running in a requestAnimationFrame loop, this will wait until the next frame.
 */
declare function nextTick(options?: CommonAwaiterOptions): CoroutineAwait<never, void>;

declare function delay<Context>(ms: number, options?: CommonAwaiterOptions): CoroutineAwait<Context, void>;

type index_NestOptions<Context, Result> = NestOptions<Context, Result>;
type index_NextAwait = NextAwait;
declare const index_all: typeof all;
declare const index_delay: typeof delay;
declare const index_nest: typeof nest;
declare const index_nextTick: typeof nextTick;
declare const index_one: typeof one;
declare namespace index {
  export {
    index_NestOptions as NestOptions,
    index_NextAwait as NextAwait,
    index_all as all,
    index_delay as delay,
    index_nest as nest,
    index_nextTick as nextTick,
    index_one as one,
  };
}

declare class FocusTargetController {
    #private;
    createManager(): FocusTargetManager;
    /**
     * Checks if the provided focus target is currently active.
     */
    isFocused(target: FocusTarget): boolean;
    isAnyFocusTargetActive(): boolean;
    /**
     * Returns `null` if there is no active focus target,
     * `undefined` if there is but it has no name set,
     * or the name if it's set.
     */
    getActiveFocusTargetDisplayName(): null | undefined | string;
}

export { AwaiterCastable, CommonAwaiterOptions, Coroutine, CoroutineAwait, CoroutineAwaitResult, CoroutineAwaitResult_Done, CoroutineAwaitResult_Wait, CoroutineContext, CoroutineController, ControllerManagerInterface as CoroutineControllerManagerInterface, CoroutineGenerator, CoroutineGeneratorFunction, CoroutineManager, ExoticCoroutineAwait, ControllerManagerInterface$1 as FocusControllerManagerInterface, FocusTarget, FocusTargetController, FocusTargetManager, FocusTargetOptions, NestOptions, NextAwait, StartCoroutineResult, index as waitUntil };
